/*
 *	This is a part of adsbox - ADS-B decoder software
 * 	(c) atty 2011-15 romulmsg@mail.ru
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>
#include "../sqlite3/sqlite3.h"
#include "misc.h"
#include "db.h"

#define CPR_EXPIRE 15 // seconds to odd/even CPR pair expire
#define AIRCRAFT_EXPIRE 60 // seconds to track expire
#define LOG_EXPIRE 600 // seconds to log expire

extern sqlite3 *db;

extern void azimuth(double *lat, double *lon, double *lat1, double *lon1,
		double *a);
extern double dist(double lat, double lon, double lat1, double lon1);
//extern double round(double value);

char *stmts[] =
		{
				"CREATE TABLE FLIGHTS (ICAO INTEGER PRIMARY KEY ASC, \
                    LAT0 INTEGER DEFAULT 0, LON0 INTEGER DEFAULT 0, LAT1 INTEGER DEFAULT 0, \
                    LON1 INTEGER DEFAULT 0, LAT REAL DEFAULT NULL, LON REAL DEFAULT NULL, \
                    LATLON_TIMESTAMP INTEGER DEFAULT ((julianday('now')-2440587.5)*86400), GND INTEGER DEFAULT NULL, \
                    ALT INTEGER DEFAULT NULL, ALT_TIMESTAMP INTEGER DEFAULT ((julianday('now')-2440587.5)*86400), \
                    CALLSIGN TEXT DEFAULT '', CAT TEXT DEFAULT '', \
                    SPEED REAL DEFAULT NULL, SPEED_INDIRECT INTEGER DEFAULT NULL, \
                    VSPEED DEFAULT NULL, VSPEED_INDIRECT DEFAULT NULL, \
                    HEADING REAL DEFAULT NULL, HEADING_INDIRECT INTEGER DEFAULT NULL, \
                    SQUAWK INTEGER DEFAULT NULL, SEEN_TIME INTEGER DEFAULT ((julianday('now')-2440587.5)*86400), \
		    APP_TIME REAL DEFAULT ((julianday('now') - 2440587.5)*86400), \
		    PARITY INTEGER DEFAULT 0, ID_SOURCE DEFAULT 0, DF17_POS INTEGER DEFAULT 0, \
		    ALT_TYPE INTEGER DEFAULT NULL, TGT_ALT INTEGER DEFAULT NULL, BARO FLOAT DEFAULT NULL, \
		    TGT_HDG FLOAT DEFAULT NULL, AP INTEGER DEFAULT NULL, VNAV INTEGER DEFAULT NULL, \
		    AHOLD INTEGER DEFAULT NULL, APPR INTEGER DEFAULT NULL, TCAS INTEGER DEFAULT NULL)",
				"CREATE INDEX IX_FLIGHTS ON FLIGHTS (ICAO)",
				"CREATE INDEX IX_FLIGHTS_ID ON FLIGHTS (ID_SOURCE)",
				"CREATE TRIGGER ON_INSERT_FLIGHTS AFTER INSERT ON FLIGHTS BEGIN \
                    INSERT INTO LOG (MSG) VALUES (\"New contact with \" || inttohex(NEW.ICAO)); END",

				"CREATE TRIGGER ON_DELETE_FLIGHTS AFTER DELETE ON FLIGHTS BEGIN \
                    DELETE FROM TRACKS WHERE ICAO=OLD.ICAO; INSERT INTO LOG (MSG) VALUES \
                    (\"Lost contact with \" || inttohex(OLD.ICAO) || \" (\" || \
                    CAST((((julianday('now') - 2440587.5)*86400) - OLD.APP_TIME)/60 AS INTEGER) || \
                    \" min)\");END",

				"CREATE TRIGGER ON_UPDATE_FLIGHTS AFTER UPDATE ON FLIGHTS BEGIN \
                    UPDATE FLIGHTS SET SEEN_TIME = (julianday('now') - 2440587.5)*86400 \
		    WHERE ICAO=OLD.ICAO; END;",
				"CREATE TRIGGER ON_UPDATE_FLIGHTS_SQUAWK AFTER UPDATE OF SQUAWK ON FLIGHTS \
		    WHEN NEW.SQUAWK != OLD.SQUAWK AND OLD.SQUAWK IS NOT NULL BEGIN \
		    INSERT INTO LOG(MSG) VALUES (inttohex(NEW.ICAO) || \": squawk has changed to \" || inttooct(NEW.SQUAWK)); END;",
				"CREATE TRIGGER ON_UPDATE_FLIGHTS_GND0 AFTER UPDATE OF GND ON FLIGHTS \
		    WHEN OLD.GND=0 AND NEW.GND=-1 BEGIN INSERT INTO LOG(MSG) VALUES (inttohex(NEW.ICAO) || \": has landed.\"); END;",
				"CREATE TRIGGER ON_UPDATE_FLIGHTS_GND1 AFTER UPDATE OF GND ON FLIGHTS \
		    WHEN OLD.GND=-1 AND NEW.GND=0 BEGIN INSERT INTO LOG(MSG) VALUES (inttohex(NEW.ICAO) || \": just airborn.\"); END;",
				"CREATE TRIGGER ON_UPDATE_FLIGHTS_TGT_ALT AFTER UPDATE OF TGT_ALT ON FLIGHTS \
		    WHEN NEW.TGT_ALT != OLD.TGT_ALT AND OLD.TGT_ALT IS NOT NULL BEGIN \
		    INSERT INTO LOG(MSG) VALUES (inttohex(NEW.ICAO) || \": target alt. has changed to \" || NEW.TGT_ALT); END;",
				"CREATE TRIGGER ON_UPDATE_FLIGHTS_AP0 AFTER UPDATE OF AP ON FLIGHTS \
		    WHEN OLD.AP=1 AND NEW.AP=0 BEGIN \
		    INSERT INTO LOG(MSG) VALUES (inttohex(NEW.ICAO) || \": switch off AP\"); END;",
				"CREATE TRIGGER ON_UPDATE_FLIGHTS_AP1 AFTER UPDATE OF AP ON FLIGHTS \
		    WHEN OLD.AP=0 AND NEW.AP=1 BEGIN \
		    INSERT INTO LOG(MSG) VALUES (inttohex(NEW.ICAO) || \": switch on AP\"); END;",

				"CREATE TRIGGER ON_UPDATE_FLIGHTS_HDG AFTER UPDATE OF LAT ON FLIGHTS \
		    WHEN OLD.LAT IS NOT NULL AND \
		    (OLD.HEADING_INDIRECT=1 OR OLD.HEADING_INDIRECT IS NULL) BEGIN \
		    UPDATE FLIGHTS SET HEADING=AZIMUTH(OLD.LAT, OLD.LON, NEW.LAT, NEW.LON), \
		    HEADING_INDIRECT=1 WHERE ICAO=OLD.ICAO; END;",

				"CREATE TRIGGER ON_UPDATE_FLIGHTS_SPEED_PRE AFTER UPDATE OF LAT ON FLIGHTS \
		    WHEN OLD.LAT IS NULL BEGIN \
		    UPDATE FLIGHTS SET LATLON_TIMESTAMP=(julianday('now') - 2440587.5)*86400 \
		    WHERE ICAO=OLD.ICAO; END;",

				"CREATE TRIGGER ON_UPDATE_FLIGHTS_SPEED AFTER UPDATE OF LAT ON FLIGHTS \
		    WHEN OLD.LAT IS NOT NULL \
		    AND DISTANCE(OLD.LAT, OLD.LON, NEW.LAT, NEW.LON) > 2 \
		    AND (OLD.SPEED_INDIRECT=1 OR OLD.SPEED_INDIRECT IS NULL) BEGIN \
		    UPDATE FLIGHTS SET SPEED=DISTANCE(OLD.LAT, OLD.LON, NEW.LAT, NEW.LON) / \
		    ((julianday('now') - 2440587.5)*86400 - OLD.LATLON_TIMESTAMP) * 3600 / 1.852, \
		    LATLON_TIMESTAMP=(julianday('now') - 2440587.5)*86400, \
		    SPEED_INDIRECT=1 WHERE ICAO=OLD.ICAO; END;",

				"CREATE TRIGGER ON_UPDATE_FLIGHTS_VSPEED AFTER UPDATE OF ALT ON FLIGHTS \
		    WHEN OLD.ALT IS NOT NULL \
		    AND ((julianday('now') - 2440587.5)*86400) - OLD.ALT_TIMESTAMP > 5 \
		    AND (OLD.VSPEED_INDIRECT=1 OR OLD.VSPEED_INDIRECT IS NULL) BEGIN \
		    UPDATE FLIGHTS SET VSPEED= \
		    CAST(ROUND((NEW.ALT-OLD.ALT) * 60 / (((julianday('now') - 2440587.5)*86400) - OLD.ALT_TIMESTAMP)) AS INTEGER), \
		    ALT_TIMESTAMP=(julianday('now') - 2440587.5)*86400, \
		    VSPEED_INDIRECT=1 WHERE ICAO=OLD.ICAO; END;",

				"CREATE TABLE RADAR_RANGE (ID INTEGER DEFAULT 0, AZIMUTH REAL DEFAULT 0, \
                    LAT DEFAULT 0, LON DEFAULT 0, DIST REAL DEFAULT 0)",
				"CREATE INDEX IX_RADAR_RANGE ON RADAR_RANGE (ID)",
				"CREATE TABLE TRACKS (ICAO INTEGER, \
                    LAT REAL DEFAULT 0, LON REAL DEFAULT 0, ALT INTEGER DEFAULT 0, \
                    SEEN_TIME INTEGER DEFAULT ((julianday('now') - 2440587.5)*86400) )",
				"CREATE INDEX IX_TRACKS ON TRACKS (ICAO)",
				"CREATE TABLE SOURCE_TCP(ID INTEGER PRIMARY KEY ASC, IP TEXT DEFAULT '', \
                    PORT INTEGER DEFAULT 30004, SHIFT INTEGER DEFAULT 0, DRIFT INTEGER DEFAULT 0, \
                    LAT REAL DEFAULT 0, LON REAL DEFAULT 0, PKTS INTEGER DEFAULT 0, BYTES INTEGER DEFAULT 0)",
				"CREATE INDEX IX_SOURCE_TCP ON SOURCE_TCP (ID)",
				"CREATE TABLE SOURCE_TCP_SBS3(ID INTEGER PRIMARY KEY ASC, IP TEXT DEFAULT '', \
				   PORT INTEGER DEFAULT 10001, SHIFT INTEGER DEFAULT 0, DRIFT INTEGER DEFAULT 0, \
				   LAT REAL DEFAULT 0, LON REAL DEFAULT 0, PKTS INTEGER DEFAULT 0, BYTES INTEGER DEFAULT 0)",
				"CREATE INDEX IX_SOURCE_TCP_SBS3 ON SOURCE_TCP_SBS3 (ID)",
				"CREATE TABLE SOURCE_TCP_BEAST(ID INTEGER PRIMARY KEY ASC, IP TEXT DEFAULT '', \
				    PORT INTEGER DEFAULT 10005, SHIFT INTEGER DEFAULT 0, DRIFT INTEGER DEFAULT 0, \
				    LAT REAL DEFAULT 0, LON REAL DEFAULT 0, PKTS INTEGER DEFAULT 0, BYTES INTEGER DEFAULT 0)",
				"CREATE INDEX IX_SOURCE_TCP_BEAST ON SOURCE_TCP_BEAST (ID)",
				"CREATE TABLE SOURCE_FILE(ID INTEGER PRIMARY KEY ASC, FILENAME TEXT DEFAULT '', \
		    PLAYRATE INTEGER DEFAULT 1, DELAY DEFAULT 0, \
                    LAT REAL DEFAULT 0, LON REAL DEFAULT 0, PKTS INTEGER DEFAULT 0, BYTES INTEGER DEFAULT 0)",
				"CREATE INDEX IX_SOURCE_FILE ON SOURCE_FILE (ID)",
				"CREATE TABLE SOURCE_SERIAL(ID INTEGER PRIMARY KEY ASC, DEVICE TEXT DEFAULT '', \
		    BAUD INTEGER DEFAULT 0, BITS DEFAULT 0, \
                    LAT REAL DEFAULT 0, LON REAL DEFAULT 0, PKTS INTEGER DEFAULT 0, BYTES INTEGER DEFAULT 0)",
				"CREATE INDEX IX_SOURCE_SERIAL ON SOURCE_SERIAL (ID)",
				"CREATE TABLE SOURCE_RTLSDR(ID INTEGER PRIMARY KEY ASC, DEVICE NUMBER DEFAULT 0, \
		    GAIN REAL DEFAULT -1, AGC INTEGER DEFAULT 0, \
                    LAT REAL DEFAULT 0, LON REAL DEFAULT 0, PKTS INTEGER DEFAULT 0, BYTES INTEGER DEFAULT 0)",
				"CREATE INDEX IX_SOURCE_RTLSDR ON SOURCE_RTLSDR (ID)",
				"CREATE TABLE SOURCE_UDP(ID INTEGER PRIMARY KEY ASC, PORT NUMBER DEFAULT 30005, \
					LAT REAL DEFAULT 0, LON REAL DEFAULT 0, PKTS INTEGER DEFAULT 0, BYTES INTEGER DEFAULT 0)",
				"CREATE INDEX IX_SOURCE_UDP ON SOURCE_UDP (ID)",
				"CREATE TABLE LOG(MSG TXT, DATE REAL DEFAULT ((julianday('now') - 2440587.5)*86400))",
				"CREATE VIEW V_SOURCES AS \
			SELECT ID, LAT, LON, PKTS, BYTES, 'sbs-3: ' || IP || ':' || PORT TXT FROM SOURCE_TCP_SBS3 UNION \
		    SELECT ID, LAT, LON, PKTS, BYTES, 'serial: ' || DEVICE TXT FROM SOURCE_SERIAL UNION \
		    SELECT ID, LAT, LON, PKTS, BYTES, 'file: ' || FILENAME TXT FROM SOURCE_FILE UNION \
		    SELECT ID, LAT, LON, PKTS, BYTES, 'tcp_avr: ' || IP || ':' || PORT TXT FROM SOURCE_TCP UNION \
			SELECT ID, LAT, LON, PKTS, BYTES, 'tcp_beast: ' || IP || ':' || PORT TXT FROM SOURCE_TCP_BEAST UNION \
		    SELECT ID, LAT, LON, PKTS, BYTES, 'rtlsdr: device ' || DEVICE TXT FROM SOURCE_RTLSDR UNION \
		    SELECT ID, LAT, LON, PKTS, BYTES, 'udp_avr: port ' || PORT TXT FROM SOURCE_UDP",
				"PRAGMA SYNCHRONOUS = OFF", "PRAGMA JOURNAL_MODE = OFF",
				NULL };

int db_delete_expired() {
	int rc;
	char *zErrMsg = 0;
	char sql[256];
	/*
	 *  Delete expired flight
	 */
	sprintf(sql,
			"DELETE FROM FLIGHTS \
        WHERE (julianday('now') - 2440587.5)*86400 - SEEN_TIME > %i",
			AIRCRAFT_EXPIRE);
	rc = sqlite3_exec(db, sql, NULL, 0, &zErrMsg);
	if (rc != SQLITE_OK) {
		DEBUG("SQLite error: %s in %s\n", zErrMsg, sql);
		sqlite3_free(zErrMsg);
		return 1;
	}
	return 0;
}

int db_delete_expired_log() {
	int rc;
	char *zErrMsg = 0;
	char sql[256];

	sprintf(sql,
			"DELETE FROM LOG \
        WHERE (julianday('now') - 2440587.5)*86400 - DATE > %i",
			LOG_EXPIRE);
	rc = sqlite3_exec(db, sql, NULL, 0, &zErrMsg);
	if (rc != SQLITE_OK) {
		DEBUG("SQLite error: %s in %s\n", zErrMsg, sql);
		sqlite3_free(zErrMsg);
		return 1;
	}
	return 0;
}

int db_update_icao(unsigned int *icao, adsb_data *data) {
	int rc;
	char *zErrMsg = 0;
	char sql[256];

	sprintf(sql, "UPDATE FLIGHTS SET ICAO=%i, ID_SOURCE=%i WHERE ICAO=%i",
			*icao, data->source_id, *icao);
	rc = sqlite3_exec(db, sql, NULL, 0, &zErrMsg);
	if (rc != SQLITE_OK) {
		DEBUG("SQLite error: %s\n %s\n", sql, zErrMsg);
		sqlite3_free(zErrMsg);
		return 1;
	}

	if (!sqlite3_changes(db)) {
		sprintf(sql, "INSERT INTO FLIGHTS (ICAO, ID_SOURCE) VALUES (%i, %i)",
				*icao, data->source_id);
		rc = sqlite3_exec(db, sql, NULL, 0, &zErrMsg);
		if (rc != SQLITE_OK) {
			DEBUG("SQLite error: %s\n %s\n", sql, zErrMsg);
			sqlite3_free(zErrMsg);
			return 1;
		}
	}

	return 0;
}

int db_update_position(unsigned int *icao, double *lat, double *lon,
		adsb_data *data) {
	int rc;
	char *zErrMsg = 0;
	char sql[256];
	double a, d;

	/*
	 *  Update record in flights table
	 */
	sprintf(sql,
			"UPDATE FLIGHTS SET LAT=%f, LON=%f, ID_SOURCE=%i, DF17_POS=1 WHERE ICAO=%i",
			*lat, *lon, data->source_id, *icao);
	rc = sqlite3_exec(db, sql, NULL, 0, &zErrMsg);
	if (rc != SQLITE_OK) {
		DEBUG("SQLite error: %s in %s\n", zErrMsg, sql);
		sqlite3_free(zErrMsg);
		return 1;
	}

	/*
	 * If nothing was updated create the record
	 */
	if (!sqlite3_changes(db)) {
		sprintf(sql,
				"INSERT INTO FLIGHTS (ICAO, LAT, LON, ID_SOURCE, DF17_POS) \
            VALUES (%i,%f,%f,%i,1)",
				*icao, *lat, *lon, data->source_id);
		rc = sqlite3_exec(db, sql, NULL, 0, &zErrMsg);
		if (rc != SQLITE_OK) {
			DEBUG("SQLite error: %s in %s\n", zErrMsg, sql);
			sqlite3_free(zErrMsg);
			return 1;
		}
	}

	/*
	 * Record the track point
	 */
	sprintf(sql,
			"INSERT INTO TRACKS (ICAO, LAT, LON, ALT) \
        SELECT %i, %f, %f, ALT FROM FLIGHTS \
        WHERE FLIGHTS.ICAO=%i AND FLIGHTS.ALT>=0",
			*icao, *lat, *lon, *icao);
	rc = sqlite3_exec(db, sql, NULL, 0, &zErrMsg);
	if (rc != SQLITE_OK) {
		DEBUG("SQLite error: %s in %s\n", zErrMsg, sql);
		sqlite3_free(zErrMsg);
		return 1;
	}

	/*
	 *  Update radar coverage area
	 */
	if ((data->lat != 0) && (data->lon != 0)) {
		azimuth(&data->lat, &data->lon, lat, lon, &a);
		d = dist(data->lat, data->lon, *lat, *lon);
		sprintf(sql,
				"UPDATE RADAR_RANGE SET \
            LAT=%f, LON=%f, DIST=%f WHERE ID=%u AND ABS(AZIMUTH - %f) < 1 AND DIST<%f",
				*lat, *lon, d, data->source_id, a, d);
		rc = sqlite3_exec(db, sql, NULL, 0, &zErrMsg);
		if (rc != SQLITE_OK) {
			DEBUG("SQLite error: %s in %s\n", zErrMsg, sql);
			sqlite3_free(zErrMsg);
			return 1;
		}
	}

	return 0;
}

int db_update_alt(unsigned int *icao, int *alt, int *gnd, adsb_data *data) {
	int rc;
	char *zErrMsg = 0;
	char sql[256];
	int rslt = 0;

	//altitude is in feets units
	if (gnd)
		sprintf(sql,
				"UPDATE FLIGHTS SET ALT=%i, GND=%i, ID_SOURCE=%i WHERE ICAO=%i",
				*alt, *gnd, data->source_id, *icao);
	else
		sprintf(sql, "UPDATE FLIGHTS SET ALT=%i, ID_SOURCE=%i WHERE ICAO=%i",
				*alt, data->source_id, *icao);

	rc = sqlite3_exec(db, sql, NULL, 0, &zErrMsg);
	if (rc != SQLITE_OK) {
		DEBUG("SQLite error: %s\n", zErrMsg);
		sqlite3_free(zErrMsg);
		return -1;
	}

	rslt = sqlite3_changes(db);

	return rslt;
}

int db_update_speed_heading(unsigned int *icao, float *speed, float *heading,
		int *vr, adsb_data * data) {
	int rc;
	char *zErrMsg = 0;
	char sql[256];

	sprintf(sql,
			"UPDATE FLIGHTS SET SPEED=%f, SPEED_INDIRECT=0, VSPEED=%i, VSPEED_INDIRECT=0, \
			HEADING=%f, HEADING_INDIRECT=0, ID_SOURCE=%i WHERE ICAO=%i",
			*speed, *vr, *heading, data->source_id, *icao);
	rc = sqlite3_exec(db, sql, NULL, 0, &zErrMsg);
	if (rc != SQLITE_OK) {
		DEBUG("SQLite error: %s\n", zErrMsg);
		sqlite3_free(zErrMsg);
		return 1;
	}

	if (!sqlite3_changes(db)) {
		sprintf(sql,
				"INSERT INTO FLIGHTS (ICAO, SPEED, SPEED_INDIRECT, VSPEED, VSPEED_INDIRECT, \
			HEADING, HEADING_INDIRECT, ID_SOURCE) \
            VALUES (%i,%f,0,%i,0,%f,0,%i)",
				*icao, *speed, *vr, *heading, data->source_id);
		rc = sqlite3_exec(db, sql, NULL, 0, &zErrMsg);
		if (rc != SQLITE_OK) {
			DEBUG("SQLite error: %s\n", zErrMsg);
			sqlite3_free(zErrMsg);
			return 1;
		}
	}

	return 0;
}

int db_update_squawk(unsigned int *icao, unsigned int *squawk, int *gnd,
		adsb_data *data) {
	int rc;
	char *zErrMsg = 0;
	char sql[256];
	int rslt = 0;

	if (gnd)
		sprintf(sql,
				"UPDATE FLIGHTS SET SQUAWK=%i, GND=%i, ID_SOURCE=%i WHERE ICAO=%i",
				*squawk, *gnd, data->source_id, *icao);
	else
		sprintf(sql, "UPDATE FLIGHTS SET SQUAWK=%i, ID_SOURCE=%i WHERE ICAO=%i",
				*squawk, data->source_id, *icao);
	rc = sqlite3_exec(db, sql, NULL, 0, &zErrMsg);
	if (rc != SQLITE_OK) {
		DEBUG("SQLite error: %s\n", zErrMsg);
		sqlite3_free(zErrMsg);
		return -1;
	}

	rslt = sqlite3_changes(db);

	return rslt;
}

int db_update_callsign(unsigned int *icao, char *callsign, char *cat,
		adsb_data *data) {
	int rc;
	char *zErrMsg = 0;
	char sql[256];

	sprintf(sql,
			"UPDATE FLIGHTS SET CALLSIGN=TRIM('%s'), CAT='%s', ID_SOURCE=%i WHERE ICAO=%i",
			(char *) callsign, (char *) cat, data->source_id, *icao);
	rc = sqlite3_exec(db, sql, NULL, 0, &zErrMsg);
	if (rc != SQLITE_OK) {
		DEBUG("SQLite error: %s\n", zErrMsg);
		sqlite3_free(zErrMsg);
		return 1;
	}

	if (!sqlite3_changes(db)) {
		sprintf(sql,
				"INSERT INTO FLIGHTS (ICAO, CALLSIGN, CAT, ID_SOURCE) \
            VALUES (%i,TRIM('%s'),'%s',%i)",
				*icao, (char *) callsign, (char*) cat, data->source_id);
		rc = sqlite3_exec(db, sql, NULL, 0, &zErrMsg);
		if (rc != SQLITE_OK) {
			DEBUG("SQLite error: %s\n", zErrMsg);
			sqlite3_free(zErrMsg);
			return 1;
		}
	}

	return 0;
}

cpr_info db_update_cpr(unsigned int *icao, unsigned char *f, int *lat, int *lon,
		int *id_source) {
	int rc;
	char *zErrMsg = 0;
	char sql[256];
	sqlite3_stmt *stmt;
	cpr_info ci;
	long int seen_time;
	time_t t = time(NULL);

	// TODO Make CPR table for each data source.

	ci.lat0 = 0;
	ci.lon0 = 0;
	ci.lat1 = 0;
	ci.lon1 = 0;
	ci.parity = 0;
	ci.lat = 0;
	ci.lon = 0;
	ci.id_source = 0;
	seen_time = t;

	sprintf(sql,
			"SELECT LAT0, LON0, LAT1, LON1, PARITY, SEEN_TIME, LAT, LON, ID_SOURCE \
        FROM FLIGHTS WHERE ICAO=%i",
			*icao);
	rc = sqlite3_prepare(db, sql, -1, &stmt, 0);
	if (rc != SQLITE_OK) {
		DEBUG("SQLite error: could not prepare.");
		sqlite3_finalize(stmt);
		return ci;
	}

	while (sqlite3_step(stmt) == SQLITE_ROW) {
		ci.lat0 = sqlite3_column_int(stmt, 0);
		ci.lon0 = sqlite3_column_int(stmt, 1);
		ci.lat1 = sqlite3_column_int(stmt, 2);
		ci.lon1 = sqlite3_column_int(stmt, 3);
		ci.parity = sqlite3_column_int(stmt, 4);
		seen_time = sqlite3_column_int(stmt, 5);
		ci.lat = sqlite3_column_double(stmt, 6);
		ci.lon = sqlite3_column_double(stmt, 7);
		ci.id_source = sqlite3_column_int(stmt, 8);
	}

	if (*f == 1) {
		ci.lat1 = *lat;
		ci.lon1 = *lon;
		if ((t - seen_time > CPR_EXPIRE) || (ci.id_source != *id_source)) {
			ci.lat0 = 0;
			ci.lon0 = 0;
		}
	} else {
		ci.lat0 = *lat;
		ci.lon0 = *lon;
		if ((t - seen_time > CPR_EXPIRE) || (ci.id_source != *id_source)) {
			ci.lat1 = 0;
			ci.lon1 = 0;
		}
	}

	sprintf(sql,
			"UPDATE FLIGHTS SET LAT0=%i, LON0=%i, LAT1=%i, LON1=%i, \
        PARITY=%i, id_source=%i WHERE ICAO=%i",
			ci.lat0, ci.lon0, ci.lat1, ci.lon1, *f, *id_source, *icao);

	rc = sqlite3_exec(db, sql, NULL, 0, &zErrMsg);
	if (rc != SQLITE_OK) {
		DEBUG("SQLite error: %s\n", zErrMsg);
		sqlite3_free(zErrMsg);
		return ci;
	}

	if (!sqlite3_changes(db)) {
		sprintf(sql,
				"INSERT INTO FLIGHTS (ICAO, LAT0, LON0, LAT1, LON1, PARITY, ID_SOURCE) VALUES \
            (%i, %i, %i, %i, %i, %i, %i)",
				*icao, ci.lat0, ci.lon0, ci.lat1, ci.lon1, *f, *id_source);

		rc = sqlite3_exec(db, sql, NULL, 0, &zErrMsg);
		if (rc != SQLITE_OK) {
			DEBUG("SQLite error: %s\n", zErrMsg);
			sqlite3_free(zErrMsg);
			return ci;
		}
	}

	sqlite3_finalize(stmt);
	return ci;
}

int db_update_source_stat(adsb_data *data) {
	int rc;
	char *zErrMsg = 0;
	char sql[512];
	int rslt = 0;

	sprintf(sql,
			"UPDATE SOURCE_TCP_SBS3 SET PKTS=PKTS+1, BYTES=%i WHERE ID=%i; \
			UPDATE SOURCE_SERIAL SET PKTS=PKTS+1, BYTES=%i WHERE ID=%i; \
			UPDATE SOURCE_FILE SET PKTS=PKTS+1, BYTES=%i WHERE ID=%i; \
			UPDATE SOURCE_TCP SET PKTS=PKTS+1, BYTES=%i WHERE ID=%i; \
			UPDATE SOURCE_TCP_BEAST SET PKTS=PKTS+1, BYTES=%i WHERE ID=%i; \
			UPDATE SOURCE_RTLSDR SET PKTS=PKTS+1, BYTES=%i WHERE ID=%i; \
			UPDATE SOURCE_UDP SET PKTS=PKTS+1, BYTES=%i WHERE ID=%i;",
			data->source_bytes, data->source_id, data->source_bytes,
			data->source_id, data->source_bytes, data->source_id,
			data->source_bytes, data->source_id, data->source_bytes,
			data->source_id, data->source_bytes, data->source_id,
			data->source_bytes, data->source_id);
	rc = sqlite3_exec(db, sql, NULL, 0, &zErrMsg);
	if (rc != SQLITE_OK) {
		DEBUG("SQLite error: %s\n", zErrMsg);
		sqlite3_free(zErrMsg);
		return -1;
	}

	rslt = sqlite3_changes(db);

	return rslt;
}

/*
 Target state and status infornation
 */
int db_update_status(unsigned int *icao, unsigned char *alt_type,
		uint16_t *tgt_alt, float *baro, float *tgt_hdg, unsigned char *ap,
		unsigned char *vnav, unsigned char *alt_hold, unsigned char *app_mode,
		unsigned char *tcas, adsb_data * data) {
	int rc;
	char *zErrMsg = 0;
	char sql[256];
	int rslt = 0;

	if (tgt_hdg)
		sprintf(sql,
				"UPDATE FLIGHTS SET ALT_TYPE=%i, TGT_ALT=%i, \
BARO=(CASE WHEN 0=%1.0f THEN NULL ELSE (%1.0f - 1) * 0.8 + 800 END), \
TGT_HDG=%f, AP=%i, VNAV=%i, AHOLD=%i, APPR=%i, TCAS=%i, ID_SOURCE=%i WHERE ICAO=%i",
				*alt_type, *tgt_alt, *baro, *baro, *tgt_hdg, *ap, *vnav,
				*alt_hold, *app_mode, *tcas, data->source_id, *icao);
	else
		sprintf(sql,
				"UPDATE FLIGHTS SET ALT_TYPE=%i, TGT_ALT=%i, \
BARO=(CASE WHEN 0=%1.0f THEN NULL ELSE (%1.0f - 1) * 0.8 + 800 END), \
AP=%i, VNAV=%i, AHOLD=%i, APPR=%i, TCAS=%i, ID_SOURCE=%i WHERE ICAO=%i",
				*alt_type, *tgt_alt, *baro, *baro, *ap, *vnav, *alt_hold,
				*app_mode, *tcas, data->source_id, *icao);

	rc = sqlite3_exec(db, sql, NULL, 0, &zErrMsg);
	if (rc != SQLITE_OK) {
		DEBUG("SQLite error: %s\n", zErrMsg);
		sqlite3_free(zErrMsg);
		return -1;
	}

	rslt = sqlite3_changes(db);

	return rslt;
}

int init_radar_range() {
	int rc;
	char sql[256], sql_s[] = "SELECT ID, LAT, LON FROM V_SOURCES";
	char *zErrMsg = 0;
	sqlite3_stmt *stmt;
	int i;

	/*
	 *  Fill with initial values
	 */

	rc = sqlite3_prepare(db, sql_s, -1, &stmt, 0);
	if (rc != SQLITE_OK) {
		DEBUG("SQLite error: cant prepare %s\n %s\n", sql_s, sqlite3_errmsg(db));
		return 1;
	}
	while (sqlite3_step(stmt) == SQLITE_ROW) {
		for (i = 0; i < 360; i = i + 2) {
			sprintf(sql,
					"INSERT INTO RADAR_RANGE(ID, AZIMUTH, LAT, LON) VALUES(%i, %i, %f, %f)",
					sqlite3_column_int(stmt, 0), i,
					sqlite3_column_double(stmt, 1),
					sqlite3_column_double(stmt, 2));
			rc = sqlite3_exec(db, sql, NULL, 0, &zErrMsg);
			if (rc != SQLITE_OK) {
				DEBUG("SQLite error: %s\n", zErrMsg);
				sqlite3_free(zErrMsg);
				return 1;
			}
		}
	}

	sqlite3_finalize(stmt);
	return 0;
}

int DEBUG_DB(char * format, ...) {
	int rc;
	char *zErrMsg = 0;
	char s[256];
	char sql[256];
	va_list args;

	va_start(args, format);
	vsnprintf(s, sizeof(s), format, args);
	va_end(args);

	snprintf(sql, sizeof(sql), "INSERT INTO LOG (MSG) VALUES (\"%s\")", s);
	rc = sqlite3_exec(db, sql, NULL, 0, &zErrMsg);
	if (rc != SQLITE_OK) {
		DEBUG("SQLite error: %s\n %s\n", sql, zErrMsg);
		sqlite3_free(zErrMsg);
		return 1;
	}

	return 0;
}

/*
 *	Twosome below are for compatibility only. Newer SQLite supports printf() function.
 */
static void f_inttohex(sqlite3_context *context, int argc, sqlite3_value **argv) {
	if (argc == 1) {
		int * t = (int *) sqlite3_value_int(argv[0]);
		char result[6];
		sprintf(result, "%06X", (unsigned int) t);
		sqlite3_result_text(context, result, -1, SQLITE_TRANSIENT);
		return;
	}
	sqlite3_result_null(context);
}

static void f_inttooct(sqlite3_context *context, int argc, sqlite3_value **argv) {
	if (argc == 1) {
		int * t = (int *) sqlite3_value_int(argv[0]);
		char result[6];
		sprintf(result, "%04o", (unsigned int) t);
		sqlite3_result_text(context, result, -1, SQLITE_TRANSIENT);
		return;
	}
	sqlite3_result_null(context);
}

static void f_azimuth(sqlite3_context *context, int argc, sqlite3_value **argv) {
	if (argc == 4) {
		double lat1 = sqlite3_value_double(argv[0]);
		double lon1 = sqlite3_value_double(argv[1]);
		double lat2 = sqlite3_value_double(argv[2]);
		double lon2 = sqlite3_value_double(argv[3]);
		double a;
		azimuth(&lat1, &lon1, &lat2, &lon2, &a);
		sqlite3_result_double(context, a);
		return;
	}
	sqlite3_result_null(context);
}

static void f_distance(sqlite3_context *context, int argc, sqlite3_value **argv) {
	if (argc == 4) {
		double lat1 = sqlite3_value_double(argv[0]);
		double lon1 = sqlite3_value_double(argv[1]);
		double lat2 = sqlite3_value_double(argv[2]);
		double lon2 = sqlite3_value_double(argv[3]);
		double d;
		d = dist(lat1, lon1, lat2, lon2);
		sqlite3_result_double(context, d);
		return;
	}
	sqlite3_result_null(context);
}

int db_init(sqlite3 *db) {
	int rc;
	char *zErrMsg = 0;
	char **stmt = stmts;

	DEBUG("Init database.\n");

	while (*stmt) {
		rc = sqlite3_exec(db, *stmt, NULL, 0, &zErrMsg);
		if (rc != SQLITE_OK) {
			DEBUG("SQLite error: %s\n", zErrMsg);
			sqlite3_free(zErrMsg);
			return 1;
		}
		stmt++;
	}

	sqlite3_create_function(db, "inttohex", 1, SQLITE_UTF8, NULL, &f_inttohex,
	NULL, NULL);
	sqlite3_create_function(db, "inttooct", 1, SQLITE_UTF8, NULL, &f_inttooct,
	NULL, NULL);
	sqlite3_create_function(db, "azimuth", 4, SQLITE_UTF8, NULL, &f_azimuth,
	NULL, NULL);
	sqlite3_create_function(db, "distance", 4, SQLITE_UTF8, NULL, &f_distance,
	NULL, NULL);
	return 0;
}
