/*
 *	adsbox - ADS-B decoder software
 *	(c) atty 2011-15 romulmsg@mail.ru
 *
 */
#include <stdio.h>
#include <getopt.h>
#include <stdlib.h> 
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <pthread.h>
#include <string.h>
#include <unistd.h>
#include <signal.h>
#include <syscall.h>
#include "misc.h"
#include "http.h"
#include "../sqlite3/sqlite3.h"
#include "adsb.h"
#include "db.h"
#include "file.h"
#include "serial.h"
#include "tcp_avr.h"
#include "tcp_sbs3.h"
#include "tcp_beast.h"
#include "rtlsdr.h"
#include "udp_avr.h"

#define SBS_PORT 30003
#define HTTP_PORT 8080

pthread_mutex_t mux;
pthread_cond_t cond_var;
char gdata[BUFLEN];
char grawdata[BUFLEN];
int dmode = 0;
_Bool record = 0;
_Bool seed_server = 0;
FILE * rd;
int delay = 0;
sqlite3 *db;

item * head;
pthread_mutex_t mux_queue;

void enqueue_data(adsb_data *d) {
	item *curr, *tail;
	
	pthread_mutex_lock(&mux_queue);

	if (head == NULL) {
		head = malloc(sizeof(item));
		memcpy(&head->list_data, d, sizeof(adsb_data));
		head->next = NULL;
	} else {
		curr = head;
		while (curr->next)
			curr = curr->next;
		tail = malloc(sizeof(item));
		curr->next = tail;
		memcpy(&tail->list_data, d, sizeof(adsb_data));
		tail->next = NULL;
	}
	pthread_mutex_unlock(&mux_queue);
}

void *decoder(void *arg) { /* decoder thread */
	unsigned char msg[128];
	struct tm *now;
	struct timespec ts;
	adsb_data *data = malloc(sizeof(adsb_data));
	item * curr;

	bzero(msg, sizeof(msg));
	while (1) {

		pthread_mutex_lock(&mux_queue);
		if (head != NULL) {
			memcpy(data, &head->list_data, sizeof(adsb_data));
			curr = head;
			head = head->next;
			free(curr);
			pthread_mutex_unlock(&mux_queue);
		} else {
			pthread_mutex_unlock(&mux_queue);
			usleep(1000);
			continue;
		}

		if (data->data_len == 7 || data->data_len == 14) {
			//DEBUG("DECODING");
			if ((data->data_len == 7 && decode_message56(data, (char *) msg))
					|| (data->data_len == 14
							&& decode_message(data, (char *) msg))) {
				bzero(grawdata, BUFLEN);
				if (!data->timestamp)
					sprintf(grawdata, "%s\r\n", data->avr_data); // raw data to seed
				else
					sprintf(grawdata, "@%012X%s\r\n", data->timestamp,
							data->avr_data + 1);
				if (record) {
					clock_gettime(CLOCK_REALTIME, &ts);
					now = localtime(&ts.tv_sec);
					fprintf(rd, "%02d:%02d:%02d.%03lu %02d.%02d.%04d %s",
							now->tm_hour, now->tm_min, now->tm_sec,
							ts.tv_nsec / 1000000, now->tm_mday, now->tm_mon + 1,
							now->tm_year + 1900, grawdata);
					fflush(rd);
				}
				pthread_mutex_lock(&mux);
				bzero(gdata, BUFLEN);
				memcpy(gdata, msg, strlen((char *) msg));
				pthread_cond_broadcast(&cond_var);
				pthread_mutex_unlock(&mux);
				bzero(msg, sizeof(msg));
				db_update_source_stat(data);
			}
		}
	}
	free(data);
	DEBUG("decoder thread exit.\n");
	pthread_exit(0);
}

void *servlet(void *arg) /* process client connection thread. SBS data seed */
{
	char buf[2];
	char cdata[BUFLEN];
	struct timespec to;

	intptr_t *s = (intptr_t*) arg;

	while (recv(*s, buf, 1, MSG_DONTWAIT)) {
		pthread_mutex_lock(&mux);
		to.tv_sec = time(NULL) + 1;
		to.tv_nsec = 0;
		pthread_cond_timedwait(&cond_var, &mux, &to);
		if (strncmp(cdata, gdata, strlen(gdata))) {
			bzero(cdata, BUFLEN);
			memcpy(cdata, gdata, strlen(gdata));
			send(*s, cdata, strlen(cdata), 0);
		}
		pthread_mutex_unlock(&mux);
	}
	close(*s);
	free(s);
	DEBUG("SBS thread exit\n");
	pthread_exit(0);
}

void *seed(void *arg) /* process client connection thread. Raw data seed */
{
	char buf[2];
	char cdata[BUFLEN];
	struct timespec to;

	intptr_t *s = (intptr_t*) arg;

	while (recv(*s, buf, 1, MSG_DONTWAIT)) {
		pthread_mutex_lock(&mux);
		to.tv_sec = time(NULL) + 1;
		to.tv_nsec = 0;
		pthread_cond_timedwait(&cond_var, &mux, &to);
		if (strncmp(cdata, grawdata, strlen(grawdata))) {
			bzero(cdata, BUFLEN);
			memcpy(cdata, grawdata, strlen(grawdata));
			send(*s, cdata, strlen(cdata), 0);
		}
		pthread_mutex_unlock(&mux);
	}
	close(*s);
	free(s);
	DEBUG("AVR seed thread exit\n");
	pthread_exit(0);
}

void usage(void) {
	printf("adsbox [options]\n\t-s, --serial input serial device\n");
	printf(
			"\t--baud serial device baudrate (valid rates are: 115200, 230400, 460800,\n");
	printf("\t\t500000, 576000, 921600, 1000000, 1152000, 1500000,\n");
	printf("\t\t2000000, 2500000, 3000000, 3500000, 4000000)\n");
	printf("\t--bits serial device bits (7 or 8)\n");
	printf(
			"\t-d, --daemon be a daemon\n\t-l, --lat data source latitude\n\t-g, --lon data source longitude\n");
	printf("\t-r, --record [file] record ADS-B data to file\n");
	printf("\t-p, --play [file] play recordered file\n");
	printf("\t-x, --rate [n] play n times faster\n");
	printf("\t-z, --snooze [n] snooze for n seconds before play\n");
	printf("\t--http-port [port] run http server on port (default 8080)\n");
	printf("\t--seed seeding raw ascii data on tcp port (default 30004)\n");
	printf("\t--seed-port [port] seed raw ascii data on tcp port\n");
	printf("\t--avr-server [IP] connect to avrascii data server\n");
	printf("\t--avr-server-port [port] connect to avrascii data server port (default 30004)\n");
	printf("\t--beast-server [IP] connect to binary beast data server\n");
	printf("\t--beast-server-port [port] connect to beast binary data server port (default 10005)\n");
#ifdef RTLSDR
	printf("\t--rtlsdr [=n] use rtlsdr device n (0 is default)\n");
	printf("\t--gain [auto|val] set rdlsdr device autogain or gain value (max available by default)\n");
	printf("\t--agc use rtlsdr device AGC\n");
#endif
	printf("\t--sbs3-server [IP] connect to SBS-3 data server\n");
	printf("\t--sbs3-server-port [port] connect to sbs3 data server port (default 10001)\n");
	printf("\t--udp-server-port [port] listen avr data on UDP port\n");
	printf("\t--udp-push-to [IP] push avrascii data to server\n");
	printf("\t--udp-push-port [port] push avrascii data to server port\n");
	printf("\t--dec-threads number of decoder threads (1 by default)\n");
	printf("\t-h, --help this help\n");
	exit(0);
}

void sigint_handler() {
	DEBUG("\nSIGINT\n");
	exit(0);
}

void long_timer() {
	db_delete_expired();
	db_delete_expired_log();
	return;
}

int main(int ac, char *av[]) {
	extern char * optarg;
	static struct option long_options[] = {
			{ "daemon", no_argument, NULL, 'd' }, { "serial",
			required_argument, NULL, 's' }, { "help", no_argument,
			NULL, 'h' }, { "lat", required_argument, NULL, 'l' }, { "lon",
			required_argument, NULL, 'g' }, { "record",
			required_argument, NULL, 'r' }, { "play",
			required_argument, NULL, 'p' }, { "rate",
			required_argument, NULL, 'x' }, { "snooze",
			required_argument, NULL, 'z' }, { "baud",
			required_argument, NULL, 1 }, { "bits", required_argument,
			NULL, 2 }, { "seed", no_argument, NULL, 3 }, { "seed-port",
			required_argument, NULL, 4 }, { "avr-server",
			required_argument, NULL, 5 }, { "avr-server-port",
			required_argument, NULL, 6 },
#ifdef RTLSDR
			{	"rtlsdr", optional_argument, NULL, 7},
			{	"gain", required_argument, NULL, 8},
			{	"agc", no_argument, NULL, 9},

#endif
			{ "sbs3-server", required_argument, NULL, 10 }, {
					"sbs3-server-port", required_argument, NULL, 11 }, {
					"dec-threads", required_argument, NULL, 12 }, {
					"udp-push-to", required_argument, NULL, 13 }, {
					"udp-push-port", required_argument, NULL, 14 }, {
					"udp-server-port", required_argument, NULL, 15 }, {
					"beast-server", required_argument, NULL, 16 }, {
					"beast-server-port",required_argument, NULL, 17 }, {
					"http-port",required_argument, NULL, 18 },
					{ 0, 0, 0, 0 } };
	int option_index;
	char s[16] = "\0";
	struct sockaddr_in addr;
	int sd, sdd = 0, max_d, en = 1;
	int pid, sid;
	char fname[64];
	double playrate = 1;
	long int baudrate = DEFAULT_SERIAL_BAUDRATE;
	long int bits = DEFAULT_SERIAL_BITS;
	int c;
	fd_set input;
	int rc;
	int seed_port = AVR_SERVER_PORT;
	unsigned int port;
	pthread_t child;
#ifdef RTLSDR
	int sdr_device = 0;
	double sdr_gain = -1;
	int sdr_agc = 0;
#endif
	unsigned int id_source = 0;
	double homelat = 0;
	double homelon = 0;
	int dec_threads = 1;

	struct sigaction sigact;
	struct sigevent sigev;
	timer_t t_id;
	struct itimerspec it;

	pthread_mutex_init(&mux, NULL);
	pthread_mutex_init(&mux_queue, NULL);
	head = NULL; // head of data queue
	udp_push_info *udp_i = NULL;
	http_info *hi = malloc(sizeof(http_info));
	hi->port = HTTP_PORT;

	bzero(fname, 64);

	DEBUG("ADSBox ADS-B decoder version %s.\n", VERSION);
	DEBUG("use --help option to see detailed help.\n");
	/*
	 * Dealing with sqlite database.
	 */
	rc = sqlite3_open(":memory:", &db);
	if (rc) {
		DEBUG("Can't open database: %s\n", sqlite3_errmsg((sqlite3 * ) &db));
		exit(1);
	}

	if (db_init(db))
		exit(1);
	/*
	 * Done with sqlite.
	 */

	while ((c = getopt_long(ac, av, "ds:hl:g:r:p:x:z:", long_options,
			&option_index)) != -1) {
		switch (c) {
		case 1:
			if (!strcmp(optarg, "115200"))
				baudrate = B115200;
			else if (!strcmp(optarg, "230400"))
				baudrate = B230400;
			else if (!strcmp(optarg, "460800"))
				baudrate = B460800;
			else if (!strcmp(optarg, "500000"))
				baudrate = B500000;
			else if (!strcmp(optarg, "576000"))
				baudrate = B576000;
			else if (!strcmp(optarg, "921600"))
				baudrate = B921600;
			else if (!strcmp(optarg, "1000000"))
				baudrate = B1000000;
			else if (!strcmp(optarg, "1152000"))
				baudrate = B1152000;
			else if (!strcmp(optarg, "1500000"))
				baudrate = B1500000;
			else if (!strcmp(optarg, "2000000"))
				baudrate = B2000000;
			else if (!strcmp(optarg, "2500000"))
				baudrate = B2500000;
			else if (!strcmp(optarg, "3000000"))
				baudrate = B3000000;
			else if (!strcmp(optarg, "3500000"))
				baudrate = B3500000;
			else if (!strcmp(optarg, "4000000"))
				baudrate = B4000000;
			else {
				DEBUG("%s - illegal baudrate.\n", optarg);
				exit(1);
			}
			db_update_source_serial_baud(&id_source, &baudrate);
			break;
		case 2:
			if (!strcmp(optarg, "7"))
				bits = CS7;
			else if (!strcmp(optarg, "8"))
				bits = CS8;
			else {
				DEBUG("%s - illegal bits option.\n", optarg);
				exit(1);
			}
			db_update_source_serial_bits(&id_source, &bits);
			break;
		case 3:
			seed_server = 1; // run seed server
			break;
		case 4:
			seed_port = atoi(optarg);
			break;
		case 5:
			id_source++;
			db_insert_source_tcp(&id_source, optarg);
			break;
		case 6:
			port = atoi(optarg);
			db_update_source_tcp_port(&id_source, &port);
			break;
#ifdef RTLSDR
			case 7:
			if (optarg != NULL) sdr_device = atoi(optarg);
			id_source++;
			db_insert_source_rtlsdr(&id_source, &sdr_device);
			break;
			case 8:
			if (!strcmp(optarg,"auto"))
			sdr_gain = -2;
			else
			sdr_gain = atof(optarg);
			if (db_update_source_rtlsdr_gain(&id_source, &sdr_gain)) {
				DEBUG("Can't set gain for source id %i\n", id_source);
				exit(1);
			}
			break;
			case 9:
			sdr_agc = 1;
			if (db_update_source_rtlsdr_agc(&id_source, &sdr_agc)) {
				DEBUG("Can't set AGC for source id %i\n", id_source);
				exit(1);
			}
			break;
#endif
		case 10:
			id_source++;
			db_insert_source_tcp_sbs3(&id_source, optarg);
			break;
		case 11:
			port = atoi(optarg);
			db_update_source_tcp_sbs3_port(&id_source, &port);
			break;
		case 12:
			dec_threads = atoi(optarg);
			break;
		case 13:
			udp_i = malloc(sizeof(udp_push_info));
			strncpy(udp_i->addr, optarg, sizeof(udp_i->addr));
			udp_i->port = 30005;
			break;
		case 14:
			udp_i->port = atoi(optarg);
			break;
		case 15:
			id_source++;
			port = atoi(optarg);
			db_insert_source_udp_avr(&id_source, &port);
			break;
		case 16:
			id_source++;
			db_insert_source_tcp_beast(&id_source, optarg);
			break;
		case 17:
			port = atoi(optarg);
			db_update_source_tcp_beast_port(&id_source, &port);
			break;
		case 18:
			hi->port = atoi(optarg);
			break;
		case 'd':
			dmode = 1;
			break;
		case 's':
			strncpy(s, optarg, 16);
			id_source++;
			db_insert_source_serial(&id_source, s);
			break;
		case 'l':
			homelat = atof(optarg);
			if (db_update_source_tcp_lat(&id_source, &homelat)
					&& db_update_source_file_lat(&id_source, &homelat)
					&& db_update_source_serial_lat(&id_source, &homelat)
#ifdef RTLSDR
					&& db_update_source_rtlsdr_lat(&id_source, &homelat)
#endif					
					&& db_update_source_tcp_sbs3_lat(&id_source, &homelat)
					&& db_update_source_tcp_beast_lat(&id_source, &homelat)
					&& db_update_source_udp_avr_lat(&id_source, &homelat)) {
				DEBUG("Can't set latitude for source id %u\n", id_source);
				exit(1);
			}
			break;
		case 'g':
			homelon = atof(optarg);
			if (db_update_source_tcp_lon(&id_source, &homelon)
					&& db_update_source_file_lon(&id_source, &homelon)
					&& db_update_source_serial_lon(&id_source, &homelon)
#ifdef RTLSDR
					&& db_update_source_rtlsdr_lon(&id_source, &homelon)
#endif
					&& db_update_source_tcp_sbs3_lon(&id_source, &homelon)
					&& db_update_source_tcp_beast_lon(&id_source, &homelon)
					&& db_update_source_udp_avr_lon(&id_source, &homelon)) {
				DEBUG("Can't set longitude for source id %u\n", id_source);
				exit(1);
			}
			break;
		case 'r':
			record = 1;
			strncpy(fname, optarg, 64);
			break;
		case 'p':
			id_source++;
			strncpy(fname, optarg, 64);
			db_insert_source_file(&id_source, fname);
			break;
		case 'x':
			playrate = atof(optarg);
			if (db_update_source_file_playrate(&id_source, &playrate)) {
				DEBUG("Can't set playrate for source id %u\n", id_source);
				exit(1);
			}
			break;
		case 'z':
			delay = atoi(optarg);
			if (db_update_source_file_delay(&id_source, &delay)) {
				DEBUG("Can't set delay for source id %u\n", id_source);
				exit(1);
			}
			break;
		case 'h':
			usage();
			break;
		case '?':
		default:
			DEBUG("Invalid option. Try --help for more info.\n");
			exit(1);
		}
	}

	signal(SIGPIPE, SIG_IGN); // ignoring SIGPIPE causing unexpected process kill
	signal(SIGINT, sigint_handler);

	if (dmode) { /* daemonize */
		printf("Entering daemon mode\n");
		pid = fork();
		if (pid < 0)
			panic("fork()");

		if (pid > 0)
			exit(0);

		umask(0);

		sid = setsid();

		if (sid < 0)
			panic("setsid()");

		close(STDIN_FILENO);
		close(STDOUT_FILENO);
		close(STDERR_FILENO);
	}

	/*
	* Long timer
	*/
	sigemptyset(&sigact.sa_mask);
	sigact.sa_flags = SA_SIGINFO;
	sigact.sa_sigaction = long_timer;

	if(sigaction(SIGUSR1, &sigact, 0) == -1)
		panic("long timer sigaction()\n");	

	sigev.sigev_notify = SIGEV_SIGNAL;
	sigev.sigev_signo = SIGUSR1;
	sigev.sigev_value.sival_int = 1;

	if (timer_create(CLOCK_REALTIME, &sigev, &t_id) == 0) {
		it.it_value.tv_sec = 60; // seconds Long timer period
		it.it_value.tv_nsec = 0;
		it.it_interval.tv_sec = it.it_value.tv_sec;
		it.it_interval.tv_nsec = it.it_value.tv_nsec;
		if (timer_settime(t_id, 0, &it, NULL) != 0)
			panic ("long timer timer_settime()\n");
	} else
		panic("long timer timer_create()\n");
	/*
	* End Long timer
	*/

	/*
	 * Start ADS-B decoder thread here
	 */
	while (dec_threads--) {
		pthread_create(&child, NULL, &decoder, NULL); /* start decoder thread */
		pthread_detach(child);
	}

	/* start UDP push thread */
	if (udp_i) {
		pthread_create(&child, NULL, &udp_avr_push, udp_i);
		pthread_detach(child);
	}

	if (record) {
		if ((rd = fopen(fname, "w+")) == NULL) {
			DEBUG("Unable to open %s\n", fname);
			exit(1);
		}
		DEBUG("Recording data to %s\n", fname);
	}

	if (init_radar_range()) {
		DEBUG("Can't init radar range\n");
		exit(1);
	}

	/* Run sources */
	run_source_file();
	run_source_tcp();
	run_source_serial();
#ifdef RTLSDR
	run_source_rtlsdr();
#endif
	run_source_tcp_sbs3();
	run_source_tcp_beast();
	run_source_udp_avr();
	
	/* Run servers */
	pthread_create(&child, NULL, &http_server, hi);
	pthread_detach(child);

	//char msg[256];
	//decode_message("8D4000FB5875444148D9A70B023A", msg);
	//exit(0);

	/*--- create sbs-1 data socket ---*/
	sd = socket(AF_INET, SOCK_STREAM, 0);
	if (sd < 0)
		panic("sbs-1 socket()");
	if (setsockopt(sd, SOL_SOCKET, SO_REUSEADDR, &en, sizeof(en)) != 0)
		panic("sbs-1 setsockopt()");

	/* non-blocking socket */
	if (ioctl(sd, FIONBIO, &en) < 0)
		panic("sbs-1 ioctl()");

	/*--- bind  ---*/
	memset(&addr, 0, sizeof(addr));
	addr.sin_family = AF_INET;
	addr.sin_port = htons(SBS_PORT);
	addr.sin_addr.s_addr = htonl(INADDR_ANY); /* any interface */
	if (bind(sd, (struct sockaddr*) &addr, sizeof(addr)) != 0)
		panic("sbs-1 bind()");
	if (listen(sd, 10) != 0)
		panic("sbs-1 listen()")
	DEBUG("Run SBS-1 data server on port %i\n", SBS_PORT);

	/*--- create seed server socket ---*/
	if (seed_server == 1) {
		sdd = socket(AF_INET, SOCK_STREAM, 0);
		if (sdd < 0)
			panic("seed socket()");
		if (setsockopt(sdd, SOL_SOCKET, SO_REUSEADDR, &en, sizeof(en)) != 0)
			panic("seed setsockopt()");

		/* non-blocking socket */
		if (ioctl(sdd, FIONBIO, &en) < 0)
			panic("seed ioctl()");

		/*--- bind  ---*/
		memset(&addr, 0, sizeof(addr));
		addr.sin_family = AF_INET;
		addr.sin_port = htons(seed_port);
		addr.sin_addr.s_addr = htonl(INADDR_ANY); /* any interface */
		if (bind(sdd, (struct sockaddr*) &addr, sizeof(addr)) != 0)
			panic("seed bind()");
		if (listen(sdd, 10) != 0)
			panic("seed listen()")
		DEBUG("Seeding avr ascii data on TCP port %i\n", seed_port);
	}

	max_d = sd;
	if ((seed_server == 1) && (sdd > max_d))
		max_d = sdd;
	max_d++;

	/*--- begin waiting for connections ---*/
	while (1) { /* process incoming clients */
		FD_ZERO(&input);
		FD_SET(sd, &input);
		if (seed_server == 1)
			FD_SET(sdd, &input);

		if (select(max_d, &input, NULL, NULL, NULL) > 0) {
			if (FD_ISSET(sd, &input)) {
				int *sd1 = (int*) malloc(sizeof(int));
				*sd1 = accept(sd, 0, 0); // accept connection
				if (pthread_create(&child, NULL, &servlet, sd1)) {
					DEBUG("pthread_create decoder error\n");
					free(sd1);
				}
				pthread_detach(child);
			}
			if (seed_server && FD_ISSET(sdd, &input)) {
				int *sdd1 = (int*) malloc(sizeof(int));
				*sdd1 = accept(sdd, 0, 0); // accept connection
				if (pthread_create(&child, NULL, &seed, sdd1)) {
					DEBUG("pthread_create seed_server error\n");
					free(sdd1);
				}
				pthread_detach(child);
			}
		}
	}
	exit(0);
}
